//
//
//
// AGC block.
//   Jeff Bilmes <bilmes@icsi.berkeley.edu>
// 
// Taken from Kohlrausch92.
// Based on some of Brian Kingsbury's code.
// 

static const char rcs_id[] = "$Header: /n/basmati/da/bedk/speech/src/frontend/RCS/Agc_Kohlrausch.cc,v 1.3 1997/11/14 19:17:29 bedk Exp bedk $";

#include "Agc_Kohlrausch.h"

#include <assert.h>
#include <math.h>

Agc_Kohlrausch::Agc_Kohlrausch(size_t iob,
			       float sf,
			       size_t nc,
			       float agctau)
  : IOBlocking(iob), samplingFrequency(sf), 
    numChannels(nc),
    // Set the filter coefficients.
    a1(exp(-(1000.0/samplingFrequency)/agctau)),
    // Set b0 for unity gain at DC.
    b0(1-a1)
{
  gains = new float[numChannels];
  result = new Matrix(numChannels,IOBlocking);
}



Agc_Kohlrausch::Agc_Kohlrausch(size_t iob,
			       float sf,
			       size_t nc,
			       float tau,
			       float db)
  : IOBlocking(iob), samplingFrequency(sf), 
    numChannels(nc),
    // Set the filter coefficients.
    a1(exp(-(1000.0/samplingFrequency)/
	   // convert db and tau over to agctau and divide.
	   (20.0*tau*M_LOG10E/db))),
    // Set b0 for unity gain at DC.
    b0(1-a1)
{
  gains = new float[numChannels];
  result = new Matrix(numChannels,IOBlocking);
}





Agc_Kohlrausch::~Agc_Kohlrausch()
{
  delete gains;
  delete result;
}

void
Agc_Kohlrausch::zeroGains()
{
  float *tmp,*tmp_ep; 
  tmp = gains;
  tmp_ep = &gains[numChannels];
  while (tmp != tmp_ep) {
    *tmp++ = 0.0;
  }
}



// The AGC signal flow is the following:
//
//  x------>[/]---------------> y 
//	     ^  	    |          
//	    g|  	    |
//	     +-[IIR LPF]----+
//			    
//  where
//      yi = xi/gi
//  and
//      g(i) = a1*g(i-1) + b0*y(i)
// 
//  corresponding to the IIR transfer function:
//
//    H(z) = b0/(1-a1z^(-1))
//
//  We call the g's the gains. 
//  We can solve for yi as a function of xi and g(i-1)
//   using the quadratic formula.
//

const CMatrix
Agc_Kohlrausch::processBlock(const CMatrix in,
			     bool& lastTime)
{
  assert(in.m == numChannels);
  assert(in.n <= IOBlocking);

  // set up a few variables for access to the input matrix.
  const size_t in_row = in.m;
  const size_t in_col = in.n;
  const size_t in_step = in.s;
  const float* inp = in.data;
  const float* const in_end = in.data + in_row * in_step;

  // variables for accessing result.
  float* res_col_p = result->data;

  // a few more locals and constants.
  const float la1 = a1;
  const float lb0 = b0;
  const float b0inv = 1/b0;

  lastTime = lastTime; // keep warning messages from occuring.  

  if (state == Start) {
    if (in.n == 0) {
      if (lastTime)
	state = End;
      return CMatrix(0,*result);
    }

    zeroGains();
    float *gainp = gains;

    while (inp != in_end) {
      const float *in_rowp = inp;
      const float *in_row_end = inp + in_col;
      float *res_row_p = res_col_p;
      float x,g;
      
      x = *in_rowp;
      if (x <= 0.0) {
	x = -sqrt(-x);
        g = -x;
      } else {
	x = sqrt(x);
        g = x;
      }
      *res_row_p++ = x;
      in_rowp++;
      while (in_rowp != in_row_end) {
	g *= la1;
	x = *in_rowp;
	if (x <= 0.0) {
	  x = -0.5*b0inv*(-g+sqrt(g*g-4.0*lb0*x));
          g -= lb0*x;
	} else {
	  x = 0.5*b0inv*(-g+sqrt(g*g+4.0*lb0*x));
          g += lb0*x;
	}
	*res_row_p++ = x;
	in_rowp++;
      }
      *gainp = g;

      inp += in_step;
      gainp++;
      res_col_p += IOBlocking;
    }
    state = Continuing;
    return CMatrix(in.n,*result);
  } else if (state == Continuing) {

    if (in.n == 0) {
      if (lastTime)
	state = End;
      return CMatrix(0,*result);
    }

    float *gainp = gains;
    while (inp != in_end) {
      const float *in_rowp = inp;
      const float *in_row_end = inp + in_col;
      float *res_row_p = res_col_p;
      float x,g;
      g = *gainp;
      while (in_rowp != in_row_end) {
	g *= la1;
	x = *in_rowp;
	if (x <= 0.0) {
	  x = -0.5*b0inv*(-g+sqrt(g*g-4.0*lb0*x));
          g -= lb0*x;
	} else {
	  x = 0.5*b0inv*(-g+sqrt(g*g+4.0*lb0*x));
          g += lb0*x;
	}
	*res_row_p++ = x;
	in_rowp++;
      }
      *gainp = g;

      inp += in_step;
      gainp++;
      res_col_p += IOBlocking;
    }

    // If this is the last time we're going to get
    // called (i.e., lastTime == true), since we're a memoryless
    // block (and don't adjust lastTime) we set our
    // state to End.
    if (lastTime)
      state = End;
    return CMatrix(in.n,*result);
  } else { // state == End
    lastTime = true;
    return CMatrix(0,*result);
  }

}


#ifdef TEST

#include <stdio.h>
#include <stdlib.h>


#define M 1
#define N 18034
#define S 19000
#define B 1500

#define MIN(x,y) ((x)<(y)?(x):(y))

#define SR 8000

main()
{
  // create an input matrix
  bool lastTime;
  Matrix inp(M,N,S);

  for (size_t i=0;i<M;i++) {
    for (size_t j=0;j<N;j++) {
      inp.data[i*S+j] = -4; // sin(2*M_PI*100*j/SR);
    }
  }

  Agc_Kohlrausch agc(B,SR,M,40.0,60.0);

  for (size_t i=0;i<N;i+=B) {
    const CMatrix cm(M,MIN(B,N-i),S,inp.data+i);
    lastTime = (bool)(i+B >= N);
    const CMatrix res = agc.processBlock(cm,lastTime);
    for (size_t j=0;j<res.n;j++) {
      printf("%f",res.data[j]);
      for (size_t k=1;k<res.m;k++) {
	printf(" %f",res.data[k*res.s+j]);
      }
      printf("\n");
    }
  }
  
  lastTime = true;
  const CMatrix cm(M,0,0,inp.data);
  const CMatrix res = agc.processBlock(cm,lastTime);
  for (size_t j=0;j<res.n;j++) {
    printf("%f",res.data[j]);
    for (size_t k=1;k<res.m;k++) {
      printf(" %f",res.data[k*res.s+j]);
    }
    printf("\n");
  }

}

#endif
