// $Header$
//
// $Log$
//
// Object to do various forms of trimming on a data stream
//

#include <stdlib.h>
#include <assert.h>

#include "Trimming.h"

const CVector
Trimming::processBlock(const CVector in,
                       bool& lastTime)
{
  assert(in.m <= ioblocking);
  assert(nc == 1);

  if ((state == End) && (points_in_buffer == 0))
    {
      lastTime = true;
      return CVector(0, buffer);
    }

  if (state == Start)
    {
      // Trim leading data
      //
      if (in.m + lead_trimmed > nLeading)
        {
          const float* src = in.data + nLeading - lead_trimmed;
          const float* const src_row_end = in.data + in.m;
          float* dest = buffer;
          while (src != src_row_end) { *dest++ = *src++; }
          points_in_buffer = in.m + lead_trimmed - nLeading;
          state = Continuing;
        }
      else
        {
          lead_trimmed += in.m;
          if (lastTime) { state = End; }
          return CVector(0, buffer);
        }
    }
  else
    {
      if (prev_out_len != 0)
        {
          // Move any remaining data to the head of the buffer
          //
          float* src = buffer + prev_out_len;
          float* src_row_end = buffer + points_in_buffer;
          float* dest = buffer;
          while (src != src_row_end) { *dest++ = *src++; }
          points_in_buffer -= prev_out_len;
        }

      if ((state != End) && (in.m != 0))
        {
          // Move input data into buffer
          //
          const float* src = in.data;
          const float* const src_row_end = in.data + in.m;
          float* dest = buffer + points_in_buffer;
          while (src != src_row_end) { *dest++ = *src++; }
          points_in_buffer += in.m;
        }
    }

  if ((state != End) && lastTime)
    {
      if (points_in_buffer > nTrailing)
        {
          points_in_buffer -= nTrailing;
        }
      else
        {
          points_in_buffer = 0;
        }
      state = End;
    }

  if (state == End)
    {
      assert(points_in_buffer <= ioblocking);
      prev_out_len = points_in_buffer;
      points_in_buffer = 0;
      return CVector(prev_out_len, buffer);
    }
  else
    {
      if (points_in_buffer < nTrailing)
        {
          prev_out_len = 0;
          return CVector(0, buffer);
        }
      else
        {
          prev_out_len = points_in_buffer - nTrailing;
          return CVector(prev_out_len, buffer);
        }
    }
}

const CMatrix
Trimming::processBlock(const CMatrix in,
                       bool& lastTime)
{
  assert(in.m == nc);

  if ((state == End) && (points_in_buffer == 0))
    {
      lastTime = true;
      return CMatrix(nc, 0, ioblocking + nTrailing, buffer);
    }

  if (state == Start)
    {
      // Trim leading data
      //
      if (in.n + lead_trimmed > nLeading)
        {
          const float* src_row_start = in.data + nLeading - lead_trimmed;
          const float* src_row_end = in.data + in.n;
          const float* src_col_end = src_row_start + in.m * in.s;
          float* dest_row_start = buffer;
          while (src_row_start != src_col_end)
            {
              const float* src = src_row_start;
              float* dest = dest_row_start;
              while (src != src_row_end) { *dest++ = *src++; }
              src_row_start += in.s;
              src_row_end += in.s;
              dest_row_start += (ioblocking + nTrailing);
            }
          points_in_buffer = in.n + lead_trimmed - nLeading;
          state = Continuing;
        }
      else
        {
          lead_trimmed += in.n;
          if (lastTime) { state = End; }
          return CMatrix(nc, 0, ioblocking + nTrailing, buffer);
        }
    }
  else
    {
      if (prev_out_len != 0)
        {
          // Move any remaining data to the head of the buffer
          //
          float* src_row_start = buffer + prev_out_len;
          float* src_row_end = buffer + points_in_buffer;
          float* src_col_end = src_row_start + nc * (ioblocking + nTrailing);
          float* dest_row_start = buffer;
          while (src_row_start != src_col_end)
            {
              float* src = src_row_start;
              float* dest = dest_row_start;
              while (src != src_row_end) { *dest++ = *src++; }
              src_row_start += (ioblocking + nTrailing);
              src_row_end += (ioblocking + nTrailing);
              dest_row_start += (ioblocking + nTrailing);
            }
          points_in_buffer -= prev_out_len;
        }

      if ((state != End) && (in.n != 0))
        {
          // Move input data into buffer
          //
          const float* src_row_start = in.data;
          const float* src_row_end = in.data + in.n;
          const float* src_col_end = src_row_start + in.m * in.s;
          float* dest_row_start = buffer + points_in_buffer;
          while (src_row_start != src_col_end)
            {
              const float* src = src_row_start;
              float* dest = dest_row_start;
              while (src != src_row_end) { *dest++ = *src++; }
              src_row_start += in.s;
              src_row_end += in.s;
              dest_row_start += (ioblocking + nTrailing);
            }
          points_in_buffer += in.n;
        }
    }

  if ((state != End) && lastTime)
    {
      if (points_in_buffer > nTrailing)
        {
          points_in_buffer -= nTrailing;
        }
      else
        {
          points_in_buffer = 0;
        }
      state = End;
    }

  if (state == End)
    {
      assert(points_in_buffer <= ioblocking);
      prev_out_len = points_in_buffer;
      points_in_buffer = 0;
      return CMatrix(nc, prev_out_len, ioblocking + nTrailing, buffer);
    }
  else
    {
      if (points_in_buffer < nTrailing)
        {
          prev_out_len = 0;
          return CMatrix(nc, 0, ioblocking + nTrailing, buffer);
        }
      else
        {
          prev_out_len = points_in_buffer - nTrailing;
          return CMatrix(nc, prev_out_len, ioblocking + nTrailing, buffer);
        }
    }
}

#ifdef TEST

#include <stdio.h>

int
main()
{
  Vector m(10);
  for (size_t i = 0; i < 10; i++) { m.data[i] = i+1; }

  for (size_t ioblocking = 1; ioblocking < 8; ioblocking++)
    {
      for (size_t ld = 0; ld < 13; ld++)
        {
          for (size_t tr = 0; tr < 13; tr++)
            {
              Trimming t(ioblocking, 1, ld, tr);

              bool EOProc = false;
              int in_offset = 0;
              int length = 0;

              printf("%d %d %d : ",ioblocking, ld, tr);

              while (!EOProc)
                {
                  if (in_offset < 10 - (int) ioblocking)
                    {
                      EOProc = false;
                      length = ioblocking;
                    }
                  else if (in_offset < 10)
                    {
                      EOProc = true;
                      length = 10 - in_offset;
                    }
                  else
                    {
                      EOProc = true;
                      length = 0;
                    }
      
                  const CVector m1(length, m.data + in_offset);
                  const CVector m2 = t.processBlock(m1, EOProc);
                  
                  for (size_t j = 0; j < m2.m; j++)
                    {
                      printf("%g ", m2.data[j]);
                    }

                  in_offset += length;
                }
              printf("\n");
            }
        }
    }
}

#endif
