/*
  fft.c -- basic radix-2, in-place FFT and IFFT routines

  All operations are on complex values, stored in linear vectors as
  re[0] im[0] re[1] im[1] ....

  $Log: fft.c,v $
  Revision 1.4  1996/08/08 19:53:18  bedk
  More efficient isPowerOf2 function

  Revision 1.3  1996/06/28 20:06:02  bedk
  made longs in basis initialization routines into size_t's

  Revision 1.2  1996/06/27 01:21:06  bedk
  Bug fix:  nextPowerOf2 was returning the log bas 2 of the answer, not the
  actual answer.

  Revision 1.1  1996/06/26 00:17:04  bedk
  Initial revision


  $Id: fft.c,v 1.4 1996/08/08 19:53:18 bedk Exp $
*/

#include <math.h>
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include "fft.h"

/* fft -- radix-2, decimation in frequency */
void
fft2t(float* x, size_t nx, const float* const w) {
  size_t x_offset = 4;
  size_t w_offset = (nx >> 1);
  float wr, wi, x0r, x0i, x1r, x1i, tr, ti;
  float* x_inner_endp;
  float* xp = x;
  const float* wp;
  const float* const x_endp = x + (nx << 1);

  /* optimize initial stage -- twiddle factor == 1 */
  while (xp != x_endp) {
    x0r = *xp;
    x0i = *(xp + 1);
    x1r = *(xp + 2);
    x1i = *(xp + 3);
    tr = x0r - x1r;
    ti = x0i - x1i;
    x0r += x1r;
    x0i += x1i;
    *xp = x0r;
    *(xp + 1) = x0i;
    *(xp + 2) = tr;
    *(xp + 3) = ti;
    xp += 4;
  }
  xp = x;

  while (x_offset < (nx << 1)) {
    while (xp != x_endp) {
      x_inner_endp = xp + x_offset;
      wp = w;
      while (xp != x_inner_endp) {
        wr = *wp;
        wi = *(wp + 1);
        x0r = *xp;
        x0i = *(xp + 1);
        x1r = *(xp + x_offset);
        x1i = *(xp + x_offset + 1);
        tr = x1r * wr - x1i * wi;
        ti = x1r * wi + x1i * wr;
        x1r = x0r - tr;
        x1i = x0i - ti;
        x0r += tr;
        x0i += ti;
        *xp = x0r;
        *(xp + 1) = x0i;
        *(xp + x_offset) = x1r;
        *(xp + x_offset + 1) = x1i;
        wp += w_offset;
        xp += 2;
      }
      xp += x_offset;
    }
    xp = x;
    w_offset >>= 1;
    x_offset <<= 1;
  }
}

/* fft -- radix-2, decimation in frequency */
void
fft2f(float* x, size_t nx, const float* const w) {
  size_t x_offset = nx;
  size_t w_offset = 2;
  float wr, wi, x0r, x0i, x1r, x1i, tr, ti;
  float* x_inner_endp;
  float* xp = x;
  const float* wp;
  const float* const x_endp = x + (nx << 1);

  while (x_offset > 2) {
    while (xp != x_endp) {
      x_inner_endp = xp + x_offset;
      wp = w;
      while (xp != x_inner_endp) {
        wr = *wp;
        wi = *(wp + 1);
        x0r = *xp;
        x0i = *(xp + 1);
        x1r = *(xp + x_offset);
        x1i = *(xp + x_offset + 1);
        tr = x0r + x1r;
        ti = x0i + x1i;
        *xp = tr;
        *(xp + 1) = ti;
        tr = x0r - x1r;
        ti = x0i - x1i;
        x1r = tr * wr - ti * wi;
        x1i = tr * wi + ti * wr;
        *(xp + x_offset) = x1r;
        *(xp + x_offset + 1) = x1i;
        wp += w_offset;
        xp += 2;
      }
      xp += x_offset;
    }
    xp = x;
    w_offset <<= 1;
    x_offset >>= 1;
  }

  /* optimize final stage -- twiddle factor == 1 */
  while (xp != x_endp) {
    x0r = *xp;
    x0i = *(xp + 1);
    x1r = *(xp + 2);
    x1i = *(xp + 3);
    tr = x0r - x1r;
    ti = x0i - x1i;
    x0r += x1r;
    x0i += x1i;
    *xp = x0r;
    *(xp + 1) = x0i;
    *(xp + 2) = tr;
    *(xp + 3) = ti;
    xp += 4;
  }
}

void
initBasis(float* x, size_t nx) {
  size_t i = 0;
  float pi, t;
  const float* const x_endp = x + nx;

  pi = 2.0 * atan2(1.0, 0.0);
  while (x != x_endp) {
    t = 2.0 * pi * i / nx;
    *x++ = cos(t);
    *x++ = -sin(t);
    i++;
  }
}

void
initInvBasis(float* x, size_t nx) {
  size_t i = 0;
  float pi, t;
  const float* const x_endp = x + nx;

  pi = 2.0 * atan2(1.0, 0.0);
  while (x != x_endp) {
    t = 2.0 * pi * i / nx;
    *x++ = cos(t);
    *x++ = sin(t);
    i++;
  }
}

/* Find log2 of a power of 2.  Dies if input is not a power of 2 */

size_t
fastLog2(size_t n) {
  size_t lg2 = 1;

  if (n == 0) {
    fprintf(stderr, "fastLog2: log2(0) = -Inf.\n");
    exit(1);
  }
  while ((n & 1) != 1) {
    n >>= 1;
    lg2++;
  }
  if (n != 1) {
    fprintf(stderr, "fastLog2: input is not a power of 2.\n");
    exit(1);
  }
  return(lg2 - 1);
}

/* Return the smallest power of 2 greater than or equal to n. */ 

size_t
nextPowerOf2(size_t n) {
  size_t lg2 = 0;
  size_t onesCount = 0;

  if (n == 0) {
    return 0;
  }
  else {
    while (n != 1) {
      lg2++;
      onesCount |= (n & 1);
      n >>= 1;
    }
    if (onesCount != 0) {
      lg2++;
    }
    return (1 << lg2);
  }
}

/* Test input to see if it is a power of 2. */

int
isPowerOf2(size_t n) {
  if (n == 0) {
    return 0;
  }
  return !(n & (n - 1));
}
