/*
   $Header: /u/spert/src/fltvec/testsuite/mul_vfmf_vf_test.c,v 1.1 1995/04/27 23:25:12 davidj Exp $

   Test mul_vfmf_vf routine.
*/

#include <stddef.h>
#include <stdlib.h>

#include "fltvec.h"
#include "rtst.h"


/* Performs one test given input values and expected output. */

void
one_test(size_t rows, size_t cols,
         const float* invec, const float* mat, const float* resvec)
{
    /* Create padded copies of inputs. */
    const size_t mat_len = rows * cols;
    float* c_invec = rtst_padvec_clone_vf(rows, invec);
    float* c_mat = rtst_padvec_clone_vf(mat_len, mat);
    float* c_resvec = rtst_padvec_new_vf(cols);

    /* Run routine. */
    mul_vfmf_vf(rows, cols, c_invec, c_mat, c_resvec);

    /* Check result. */
    rtst_checkeq_vfvf(cols, c_resvec, resvec);

    /* Check input not corrupted. */
    rtst_checkeq_vfvf(rows, c_invec, invec);
    rtst_checkeq_vfvf(mat_len, c_mat, mat);

    /* Free and check padding. */
    rtst_padvec_del_vf(c_invec);
    rtst_padvec_del_vf(c_mat);
    rtst_padvec_del_vf(c_resvec);
}

void
test_1x1(void)
{
    static const float v[] = {  5.0f };
    static const float m[] = { 12.0f };
    static const float r[] = { 60.0f };

    rtst_log("Running test_1x1().\n");
    one_test(1, 1, v, m, r);
    rtst_log("Passed test_1x1().\n");
}

void
test_3x2(void)
{
    static const float v[] = { 7.0f, -5.0f, +4.0f };
    static const float m[] = {
        1.0f,  3.0f,
        -2.0f,  9.0f,
        3.0f, -3.0f
    };
    static const float r[] = { 29.0f, -36.0f };

    rtst_log("Running test_3x2().\n");
    one_test(3, 2, v, m, r);
    rtst_log("Passed test_3x2().\n");
}

void
run_directed_tests(void)
{
    rtst_log("\nRunning directed tests.\n");
    test_1x1();
    test_3x2();
    rtst_log("Passed directed tests.\n\n");
}

void
run_random_tests(void)
{
    const int numtests = rtst_numtests;
    const size_t maxlen = rtst_sizetests;
    float *invec = malloc_vf(maxlen);
    float *mat = malloc_vf(maxlen*maxlen);
    float *outvec = malloc_vf(maxlen);
    int i;

    rtst_log("\nStarting random tests against naive code.\n");

    for (i=0; i<numtests; i++)
    {
        size_t rows = (size_t) rtst_urand_i32i32_i32(1, maxlen);
        size_t cols = (size_t) rtst_urand_i32i32_i32(1, maxlen);

        rtst_urandi_i32i32_vf(rows, -7, +7, invec);
        rtst_urandi_i32i32_vf(rows*cols, -7, +7, mat);

        nv_mul_vfmf_vf(rows, cols, invec, mat, outvec);

        rtst_log("rows=%d cols=%d\n", rows, cols);

        one_test(rows, cols, invec, mat, outvec);
    }

    rtst_log("Passed random tests against naive code.\n\n");

    free_vf(invec);
    free_vf(mat);
    free_vf(outvec);
}

int
main(int argc, char *argv[])
{
    rtst_args(argc, argv);

    rtst_start("mul_vfmf_vf");
    run_directed_tests();
    run_random_tests();
    rtst_passed();
    
    rtst_exit();

    return 0;
}
