// -*- C++ -*-
// $Header: /n/abbott/da/drspeech/src/quicknet/RCS/QN_MLPWeightFile_MLPW.h,v 1.2 2001/03/13 19:56:57 dpwe Exp $
// 
// QN_MLPWeightFle_MLPW.h
//
// Define a new *binary* MLP Weight file format
// 1997jun19 dpwe@icsi.berkeley.edu
//

/* 2001-02-17 dpwe@ee.columbia.edu
 * STATUS:
 * I'm probably going to have to leave this for a little, so here's 
 * where I am.  QN_MLPWeightsFile_MLPW works for both reading 
 * and writing, and has been tested for binary float and short 
 * representations.  (The short representation, 16 bit/wt, is about 1/4 
 * the size of the ascii RAP3 format and does not signficantly reduce 
 * accuracy, since we have used 16 bit weights on the SPERT, and the 
 * representation is the same).
 * qncopywts is a program to support conversion between different 
 * weights files formats. 
 *   However, I think the file format needs to be modified.  It turns out 
 * that bias weights occupy a rather different range from other weights - 
 * in the region of -log(inputsize) or something, so they typically need 
 * a different fixed-point representation.  On the SPERT, this is handled 
 * by having different 'shifts' (exponents) for the bias weights, and in 
 * fact these shifts can vary by layer.  So I think the right way to have 
 * fixed-point wts files would be to have the exponent stored separately 
 * for each section, to permit greatest flexibility in storage.  This 
 * makes the current 'weightScale' parameter superfluous.  So, for
 * instance, we could have an nSections parameter in the header, 
 * then an array of sectionExps (but only for integer datatypes?).
 *   I also haven't implemented the netType flag yet.
 *   Then in due course we need to add weights_file_format flags to 
 * qnstrn and qnsfwd to allow reading and writing.  Eventually, I would 
 * imagine having MLPW_F or MLPW_S being the standard net type accepted 
 * by ffwd too.
 */

#ifndef QN_MLPWeightFile_MLPW_h_INCLUDED
#define QN_MLPWeightFile_MLPW_h_INCLUDED


/* Must include the config.h file first */
#include <QN_config.h>
#include <stdio.h>
#include "QN_types.h"
#include "QN_MLPWeightFile.h"

enum QN_MLPW_dtype {
    QN_MLPW_dt_none = 0,
    QN_MLPW_dt_int8 = 1, 
    QN_MLPW_dt_int16 = 2, 
    QN_MLPW_dt_int32 = 4, 
    QN_MLPW_dt_float = (4+32), 
    QN_MLPW_dt_double = (8+32)
};

// Net types
enum QN_MLPW_ntype {
    QN_MLPW_nt_none = 0,
    QN_MLPW_nt_softmax,
    QN_MLPW_nt_sigmoid,
    QN_MLPW_nt_linear
};

// Convert a dtype constant into bytes-per-value
#define QN_MLPW_dt_sizeof(type) (type & 0x0F)

// Magic number at start of file
#define QN_MLPW_MAGIC "MLPW"
#define QN_MLPW_SIZEOFMAGIC 4

#define QN_MLPW_VERSION (20010313)

// Limit so we can use a fixed array for nlayers
#define QN_MLPW_MAX_NLAYERS	4
#define QN_MLPW_MAX_NSECTS	2*(QN_MLPW_MAX_NLAYERS-1)

/* Header structure for MLPW files */
typedef struct _MLPW_hdr {
    QNUInt32 	magic;		// "MLPW" magic number
    QNUInt32 	version;	// version field
    QNUInt32	nettype;	// flag indicating nettype (softmax etc)
    QNUInt32	nlayers;	// total number of layers (1 hidden -> 3)
    QNUInt32	nsections;	// total number of sections (MLP3 = 4)
    // On disk, this header is followed by:
    // QNUInt32   layersize[nlayers]      number of units in each layer
    // then the (4) sections...
} QN_MLPW_hdr_str;

typedef struct _MLPW_sect_hdr {
    // This is the start of each section
    QNUInt32	sectiontype;	// QN_SectionSelector value
    QNUInt32	numvalues;	// how many values
    QNUInt32	datatype;	// QN_MLPW_dt_* field
    // then, for fixed-point datatypes, the exponent in 4 bytes
    // then the coded values
} QN_MLPW_sect_hdr;

class QN_MLPWeightFile_MLPW : public QN_MLPWeightFile
{
  public:
    // If n_layers is nonzero for read, it is checked against the file
    QN_MLPWeightFile_MLPW(int dbg, FILE* stream, QN_FileMode mode,
			  const char* name,
			  int n_layers, int layer_sizes[], 
			  QN_MLPW_ntype nt = QN_MLPW_nt_none);

    // How many layers are there?
    size_t num_layers() { return nLayers; }

    // How many sections?
    size_t num_sections() { return nSections; }

    // How many units are there in the given layer?
    size_t size_layer(QN_LayerSelector layer) { assert(layer < nLayers); 
                                                return layerSize[layer]; }

    // Return whether weight matrices are input or output major
    enum QN_WeightMaj get_weightmaj(QN_SectionSelector);

    // Return the order of the weights in the file
    enum QN_SectionSelector get_weighttype(int section);

    // Is this fine QN_READ or QN_WRITE?
    enum QN_FileMode get_filemode();

    // Set the fixed-point exponents
    void set_datatype(QN_SectionSelector sect, QN_MLPW_dtype dt, 
		      int exp = 0);  /* exp for fx types only */

    void get_datatype(QN_SectionSelector sect, QN_MLPW_dtype *dt, 
		      int *exp = NULL);  /* exp for fx types only */

    // Read in a buffer full of weights
    size_t read(float* buf, size_t count);

    // Write out a buffer full of weights
    size_t write(const float* buf, size_t count);
    
private:
    int netType;	// softmax etc
    int nLayers;
    int layerSize[QN_MLPW_MAX_NLAYERS];
    int nSections;

    // data types and optional fixed-point exponents for each layer
    QN_MLPW_dtype datatype[QN_MLPW_MAX_NSECTS];
    int exponents[QN_MLPW_MAX_NSECTS];
    QN_MLPW_dtype dtype;		// prevailing data type
    int exp;		// prevailing exp, passed to vx conv routines

    int io_count;
    int io_state;

private:
    int debug;			// Level of verbosity
    const enum QN_FileMode mode;
    FILE* const stream;		// Stream used to access weights files
    const char* filename;	// Name of the weight file

    static const char* CLASSNAME; // Name of this class


    int read_header(void);
    int write_header(void);
    int read_sect_header(int sect, 
			 int numvals, 
			 QN_MLPW_dtype* datatype, 
			 int* exponent);
    int write_sect_header(int sect, 
			  int numvals, 
			  QN_MLPW_dtype datatype, 
			  int exponent /* = 0 */);
};

#endif

