/* Must include the config.h file first */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdlib.h>
#include <assert.h>
#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include "QuickNet.h"
#include "rprf.h"
#include "fltvec.h"
#include "IPM_timer.h"

#ifndef EXIT_FAILURE
#define EXIT_FAILURE 1
#define EXIT_SUCCESS 0
#endif

enum
{
    DEFAULT_BUF_FRAMES = 16
};
enum QN_OutputLayerType OUT_TYPE = QN_OUTPUT_SOFTMAX;

int
main(int argc, const char** argv)
{
    size_t n_input;
    size_t n_hidden;
    size_t n_output;
    size_t bunch_size;
    const char* opts = "";

    size_t buf_frames;
    float* input_buf;
    float* output_buf;
    float* target_buf;

    // Usage: MLP3_perf <n_in> <n_hid> <n_out> <bunch> <opts>
    // "1" -> 16 bit
    // "x" -> fixed point
    // "b" -> bunch mode
    // "f" -> forward pass
    rprf_parse_args(argc, argv, "iiiis", &n_input, &n_hidden, &n_output,
		    &bunch_size, &opts);
    assert(n_input>0 && n_hidden>0 && n_output>0);
    assert(bunch_size>=1);

    buf_frames = bunch_size;
    input_buf = malloc_vf(n_input * buf_frames);
    output_buf = malloc_vf(n_output * buf_frames);
    target_buf = malloc_vf(n_output * buf_frames);
    copy_f_vf(n_input*buf_frames, 0.0f, input_buf);
    copy_f_vf(n_output*buf_frames, 0.0f, output_buf);
    copy_f_vf(n_output*buf_frames, 0.0f, target_buf);

    size_t n_conns = (n_input + n_output) * n_hidden * buf_frames;

    QN_MLP* net;
    if (strchr(opts, 'x'))
    {
#ifdef QN_HAVE_LIBFXLIB
	if (!strchr(opts, '1'))
	{
	    // Online fixed point, 32 bit weights.
	    net = new QN_MLP_Online1632Fx3(0, n_input, n_hidden,
					   n_output,  OUT_TYPE);
	}
	else
	{
	    // Online fixed point, 16 bit weights.
	    net = new QN_MLP_OnlineFx3(0, n_input, n_hidden, n_output,
				       OUT_TYPE);
	}
#else
	fprintf(stderr, "no fixed point support in this version "
		"of QuickNet.\n");
	net = NULL;
	exit(EXIT_FAILURE);
#endif
    }
    else
    {
	if (strchr(opts, 'b'))
	{
	    // Online bunch mode.
	    net = new QN_MLP_BunchFl3(0, "net", n_input, n_hidden, n_output,
				      OUT_TYPE, bunch_size);
	}
	else
	{
	    // Online floating point.
	    net = new QN_MLP_OnlineFl3(0, "net", n_input, n_hidden, n_output,
				       OUT_TYPE);
	}
    }

    net->set_learnrate(QN_MLP3_INPUT2HIDDEN, 0.001);
    net->set_learnrate(QN_MLP3_HIDDEN2OUTPUT, 0.001);
    net->set_learnrate(QN_MLP3_HIDDENBIAS, 0.001);
    net->set_learnrate(QN_MLP3_OUTPUTBIAS, 0.001);

#define FORWARD net->forward(buf_frames, input_buf, output_buf)
#define TRAIN net->train(buf_frames, input_buf, target_buf, output_buf)

    if (strchr(opts, 'f'))
    {
	RPRF_BODY(FORWARD, n_conns);
    }
    else
    {
	RPRF_BODY(TRAIN, n_conns);
    }

    delete net;

    exit(EXIT_SUCCESS);
}
